#include "mplayermediaidentifier.h"

#include "mplayerbackendaddon.h"

#include <cstdio>
#include <cstdlib>
#include <unistd.h>

#ifdef QT4
#include <QProcess>
#include <QHash>
#else
#include <qtimer.h>
#include <qprocess.h>
#include <qdict.h>
#endif

#include "debug.h"
#include "helpers.h"

#include "compathack.h"

MPlayerMediaIdentifier::MPlayerMediaIdentifier(MPlayerConfiguration &config)
    :	config_(config),
        media_(NULL),
        fileScanningFinished_(false),
        fileWasIdentified_(false),
        identifyProcess_(NULL),
        tempLine_("")
{
    connect(&killIdentifyProcessTimer_, SIGNAL(timeout()), this, SLOT(killIdentifyProcessTimerTimedOut()));
}

MPlayerMediaIdentifier::~MPlayerMediaIdentifier()
{
    DGUARDMETHODTIMED;
    DPRINTF("Killing mplayer slave because we're quitting...");
    cleanUpIdentifyProcess();
}

void MPlayerMediaIdentifier::cleanUpIdentifyProcess()
{
    if (identifyProcess_)
    {
#ifdef QT4
        identifyProcess_->terminate();
#else
        identifyProcess_->tryTerminate();
#endif

        usleep(1000000); // wait 1 sec
#ifdef QT4
        if (identifyProcess_->state() != QProcess::NotRunning)
#else
        if (identifyProcess_->isRunning())
#endif
        {
            identifyProcess_->kill();
            usleep(500000); // wait 500 ms
        }

        delete identifyProcess_;
        identifyProcess_ = NULL;
    }
}

void MPlayerMediaIdentifier::killIdentifyProcessTimerTimedOut()
{
    DGUARDMETHODTIMED;
    killIdentifyProcessTimer_.stop();
    DPRINTF("Killing mplayer slave since we no longer need it...");
    cleanUpIdentifyProcess();
}

void MPlayerMediaIdentifier::sendCommand(const QString &command)
{
    writeToProcess(identifyProcess_, command);
}

bool MPlayerMediaIdentifier::identify(Media *media, const MediaLocation &location)
{
    DGUARDMETHODFTIMED("location: %s", (const char *)location.toString().toUtf8());

    // kill mplayer slave after 10 seconds, if no other identification request
    // happens within that period...
    killIdentifyProcessTimer_.setInterval(10000);

    // perhaps we are retrying after a timeout?
    // when this happens, media_ is still set, so use this information to
    // recognize this state...
    bool mediaWasSet = media_;

    media_ = media;
    fileScanningFinished_ = false;
    fileWasIdentified_ = false;

#ifdef QT4
    if (!identifyProcess_ || identifyProcess_->state() == QProcess::NotRunning)
#else
    if (!identifyProcess_ || !identifyProcess_->isRunning())
#endif
    {
        QStringList args;
#ifndef QT4
        args << config_.mplayerBinLocation();
#endif
        args << "-vo" << "null" << "-ao" << "null" << "-v" << "-nocache" << "-slave" << "-identify";
        args << location.toString();

        DPRINTF("Starting mplayer slave...");
#ifdef QT4
        identifyProcess_ = new QProcess();
        identifyProcess_->setProcessChannelMode(QProcess::MergedChannels);
        identifyProcess_->setReadChannel(QProcess::StandardOutput);
        identifyProcess_->start(config_.mplayerBinLocation(), args);
        identifyProcess_->waitForStarted(3000);
#else
        identifyProcess_ = new QProcess(args);
        identifyProcess_->start();
#endif

        DPRINTF("Sending pause...");
        sendCommand("pause");
    }
    else
    {
        DPRINTF("Mplayer slave still running, inputing new file...");
        sendCommand("pausing loadfile \"" + location.toString() + "\" 0");
    }

    DPRINTF("Waiting for file scanning to finish...");

    // timeout timer...
    QTime time;
    time.start();
    bool timedOut = false;

#ifdef QT4
    while ((identifyProcess_->state() != QProcess::NotRunning || identifyProcess_->canReadLine()) && !fileScanningFinished_ && !timedOut)
#else
    while ((identifyProcess_->isRunning() || identifyProcess_->canReadLineStdout()) && !fileScanningFinished_ && !timedOut)
#endif
    {
        timedOut = time.elapsed() > 5000;
        parseLines();
#ifdef QT4
        identifyProcess_->waitForReadyRead(1);
#else
        usleep(1000); // sleep 1 ms
#endif
    }

    if (timedOut)
    {
        DPRINTF("Timeout while scanning file...");
        cleanUpIdentifyProcess();

        if (!mediaWasSet)
        {
            DPRINTF("Retrying again...");
            return identify(media, location);
        }
    }

    media_ = NULL;
    return fileWasIdentified_;
}

void MPlayerMediaIdentifier::parseLines()
{
    DGUARDMETHOD;

    if (!media_)
        return;

#ifdef QT4
    while (identifyProcess_->canReadLine())
    {
        QString line = identifyProcess_->readLine();
#else
    while (identifyProcess_->canReadLineStdout())
    {
        QString line = identifyProcess_->readLineStdout();
#endif

        DPRINTF("line: %s", (const char *)line.toUtf8());

        if (line.startsWith("ID_"))
        {
            // ID_*
            line = line.mid(3);
            if (line.startsWith("LENGTH="))
            {
                line = line.mid(7);
                int posDot = line.indexOf(".");
                media_->setLength(posDot > -1 ? line.left(posDot).toInt() : line.toInt());
                fileWasIdentified_ = true;
            }
            else if (line.startsWith("AUDIO_"))
            {
                // ID_AUDIO_*
                line = line.mid(6);

                // avoid stepping down from video to audio...
                if (media_->mediaType() == Media::MEDIATYPE_UNKNOWN)
                    media_->setMediaType(Media::MEDIATYPE_AUDIO);

                if (line.startsWith("CODEC="))
                    media_->mediaAudio()->codec = line.mid(6);
                else if (line.startsWith("BITRATE="))
                    media_->mediaAudio()->bitrate = line.mid(8).toInt();
                else if (line.startsWith("RATE="))
                    media_->mediaAudio()->sampleRate = line.mid(5).toInt();
                else if (line.startsWith("NCH="))
                    media_->mediaAudio()->channels = line.mid(4).toInt();

                fileWasIdentified_ = true;
            }
            else if (line.startsWith("VIDEO_"))
            {
                // ID_VIDEO_*
                line = line.mid(6);

                media_->setMediaType(Media::MEDIATYPE_VIDEO);

                if (line.startsWith("FORMAT="))
                    media_->mediaVideo()->codec = line.mid(7);
                else if (line.startsWith("BITRATE="))
                    media_->mediaVideo()->bitrate = line.mid(8).toInt();
                else if (line.startsWith("WIDTH="))
                    media_->mediaVideo()->width = line.mid(6).toInt();
                else if (line.startsWith("HEIGHT="))
                    media_->mediaVideo()->height = line.mid(7).toInt();
                else if (line.startsWith("FPS="))
                    media_->mediaVideo()->FPS = line.mid(4).toFloat();
                else if (line.startsWith("ASPECT="))
                    media_->mediaVideo()->aspectRatio = line.mid(7).toFloat();

                fileWasIdentified_ = true;
            }
            else if (line.startsWith("CLIP_INFO_NAME"))
            {
                tempLine_ = "";

                int posEq = line.indexOf('=');
                if (posEq > -1)
                    tempLine_ = line.mid(posEq + 1);
            }
            else if (line.startsWith("CLIP_INFO_VALUE") && tempLine_.length() > 0)
            {
                int posEq = line.indexOf('=');
                if (posEq > -1)
                {
                    line = line.mid(posEq + 1);

                    if (containsUTF8(line))
                        line = reinterpretUTF8(line);

                    media_->mediaMetadataExt()->addMediadata(tempLine_, line);

                    fileWasIdentified_ = true;
                }
            }
        }
        else if (line.startsWith("OggVorbisComment: "))
        {
            media_->setMediaType(Media::MEDIATYPE_AUDIO);
            line = line.mid(18);
            int sepIndex = line.indexOf('=');
            if (sepIndex > 0)
            {
                QString tag(line.left(sepIndex));
                QString value(line.mid(sepIndex + 1));

                if (containsUTF8(value))
                    value = reinterpretUTF8(value);

                media_->mediaMetadataExt()->addMediadata(tag, value);

                fileWasIdentified_ = true;
            }
        }
        else if (line.startsWith("SLAVE: Couldn't open demuxer"))
        {
            media_->setMediaType(Media::MEDIATYPE_UNKNOWN);
        }
        else if (line.startsWith("Starting playback"))
        {
            //qDebug("Got finish signal...");
            fileScanningFinished_ = true;
        }
    }
}
